

set -ex



pip check
ipcluster --help
ipcontroller --help
ipengine --help
jupyter labextension list
jupyter server extension list
jupyter serverextension list
jupyter labextension list 1>labextensions 2>&1
jupyter server extension list 1>server_extensions 2>&1
jupyter serverextension list 1>serverextensions 2>&1
grep -ie "ipyparallel-labextension.*v8\.4\.1.*OK" labextensions
grep -ie "ipyparallel.*8\.4\.1.*OK" server_extensions
grep -ie "ipyparallel.*8\.4\.1.*OK" serverextensions
exit 0
