#  tests for ipyparallel-8.4.1-py39hd43f75c_0 (this is a generated file);
print('===== testing package: ipyparallel-8.4.1-py39hd43f75c_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
#!/usr/bin/env python

import ipyparallel
import ipyparallel.apps
import ipyparallel.client
import ipyparallel.controller
import ipyparallel.engine
import ipyparallel.serialize

import sys
from subprocess import check_output, STDOUT

def sh(cmd):
    out = check_output(cmd, stderr=STDOUT)
    return out.decode('utf8', 'replace')

# verify extension enabling
nbextensions = sh([sys.executable, '-m', 'jupyter', 'nbextension', 'list'])
print(nbextensions)
assert 'ipyparallel' in nbextensions

serverextensions = sh([sys.executable, '-m', 'jupyter', 'serverextension', 'list'])
print(serverextensions)
assert 'ipyparallel' in serverextensions
#  --- run_test.py (end) ---

print('===== ipyparallel-8.4.1-py39hd43f75c_0 OK =====');
